﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;

namespace Socks5to4a
{
    /// <summary>
    /// Socks5to4a creates a SOCKS 5 proxy interface on localhost
    /// that forwards requests to a SOCKS 4a proxy on the network.
    /// </summary>
    /// <remarks>
    /// References:
    /// http://en.wikipedia.org/wiki/SOCKS
    /// http://ftp.icm.edu.pl/packages/socks/socks4/SOCKS4.protocol
    /// http://tools.ietf.org/rfc/rfc1928.txt
    /// </remarks>
    class Socks5to4a
    {
        /// <summary>
        /// Local port on which to listen for SOCKS 5 traffic.
        /// </summary>
        private const int ListenPort = 1080;

        /// <summary>
        /// Program entry point.
        /// </summary>
        /// <param name="args">Arguments.</param>
        [SuppressMessage("Microsoft.Usage", "CA1806:DoNotIgnoreMethodResults", MessageId = "Socks5to4a.NetworkStreamRelay",
            Justification = "NetworkStreamRelay manages itself.")]
        public static void Main(string[] args)
        {
            try
            {
                // Process command-line arguments
                int socks4aProxyPort;
                if (((args.Length < 2) || (3 < args.Length)) || !int.TryParse(args[1], out socks4aProxyPort))
                {
                    throw new NotSupportedException("SYNTAX: Socks5to4a.exe Socks4aProxyName Socks4aProxyPort [FallbackHostName]");
                }
                var socks4aProxyName = args[0];
                var fallbackHostName = (3 == args.Length) ? args[2] : null;

                // Listen for incoming connections on localhost
                var nextConnectionId = 1;
                var listener = new TcpListener(IPAddress.Loopback, ListenPort);
                try
                {
                    Log(0, "Listening on localhost:{0}{1}...", ListenPort,
                        (null != fallbackHostName) ? string.Format(CultureInfo.CurrentCulture, " with fall-back to {0}", fallbackHostName) : "");
                    listener.Start();
                    while (true)
                    {
                        var connectionId = nextConnectionId++;
                        Socks5Server socks5Server = null;
                        Socks4Client socks4Proxy = null;
                        try
                        {
                            // Create SOCKS 5 server to handle connection
                            socks5Server = new Socks5Server(listener.AcceptTcpClient());
                            Log(connectionId, "Accepted connection.");

                            // Read client request
                            Log(connectionId, "Reading client request...");
                            socks5Server.Authenticate();
                            IEnumerable<byte> host;
                            IEnumerable<byte> port;
                            bool hostIsDomain;
                            socks5Server.ReadConnectRequest(out host, out port, out hostIsDomain);

                            // Display client request
                            var hostString = hostIsDomain ?
                                Encoding.ASCII.GetString(host.ToArray()) :
                                string.Join(".", host.Select(b => b.ToString(CultureInfo.InvariantCulture)).ToArray());
                            var portString = BitConverter.ToUInt16(port.Reverse().ToArray(), 0);
                            Log(connectionId, "Client requests {0}:{1}.", hostString, portString);

                            // Connect to SOCKS 4a proxy
                            Log(connectionId, "Connecting to proxy...");
                            socks4Proxy = new Socks4Client(socks4aProxyName, socks4aProxyPort);
                            Log(connectionId, "Connected.");

                            // Pass request on to SOCKS 4a proxy
                            Log(connectionId, "Forwarding request...");
                            var connected = socks4Proxy.Connect(host, port, hostIsDomain);

                            // Handle connect failure by retrying with fallback (if present)
                            if (!connected && (null != fallbackHostName))
                            {
                                Log(connectionId, "Connect failed.");
                                Log(connectionId, "Retrying with {0}:{1}...", fallbackHostName, portString);
                                host = Encoding.ASCII.GetBytes(fallbackHostName);
                                hostIsDomain = true;
                                socks4Proxy = new Socks4Client(socks4aProxyName, socks4aProxyPort);
                                connected = socks4Proxy.Connect(host, port, hostIsDomain);
                            }

                            if (connected)
                            {
                                // Send success response
                                Log(connectionId, "Connected.");
                                Log(connectionId, "Sending response...");
                                socks5Server.SendConnectResponse(true, host, port, hostIsDomain);

                                // Proxy data
                                Log(connectionId, "Proxying data...");
                                new NetworkStreamRelay(socks5Server.ProxiedStream, socks4Proxy.ProxiedStream,
                                    () => Log(connectionId, "Connection closed."));
                            }
                            else
                            {
                                // Send failure response
                                Log(connectionId, "Connect failed.");
                                Log(connectionId, "Sending response...");
                                socks5Server.SendConnectResponse(false, host, port, hostIsDomain);

                                // Close connections
                                Log(connectionId, "Closing connection.");
                                socks4Proxy.Close();
                                socks5Server.Close();
                            }
                        }
                        catch (IOException)
                        {
                            Log(connectionId, "I/O exception; closing connection.");
                            if (null != socks4Proxy)
                            {
                                socks4Proxy.Close();
                            }
                            if (null != socks5Server)
                            {
                                socks5Server.Close();
                            }
                        }
                    }
                }
                finally
                {
                    // Stop listening
                    listener.Stop();
                }
            }
            catch (Exception e)
            {
                // Output exception information
                Console.WriteLine(e);
            }
        }

        /// <summary>
        /// Log a message to the console.
        /// </summary>
        /// <param name="connectionId">Connection ID.</param>
        /// <param name="format">Format string.</param>
        /// <param name="args">Params.</param>
        private static void Log(int connectionId, string format, params object[] args)
        {
            Console.WriteLine("{0}: {1}", connectionId, string.Format(CultureInfo.CurrentCulture, format, args));
        }
    }

    /// <summary>
    /// Base class for SOCKS client/server.
    /// </summary>
    class SocksBase
    {
        protected TcpClient _tcpClient;
        protected NetworkStream _stream;

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="tcpClient">TcpClient instance.</param>
        public SocksBase(TcpClient tcpClient)
        {
            _tcpClient = tcpClient;
            _stream = _tcpClient.GetStream();
        }

        /// <summary>
        /// Proxied NetworkStream instance.
        /// </summary>
        public NetworkStream ProxiedStream
        {
            get { return _stream; }
        }

        /// <summary>
        /// Closes the stream/connection.
        /// </summary>
        public void Close()
        {
            _stream.Close();
            _tcpClient.Close();
        }
    }

    /// <summary>
    /// Implements simple SOCKS 5 server behavior for a single connection.
    /// </summary>
    class Socks5Server : SocksBase
    {
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="client">TcpClient instance.</param>
        public Socks5Server(TcpClient client)
            : base(client)
        {
        }

        /// <summary>
        /// Processes the authentication request.
        /// </summary>
        public void Authenticate()
        {
            // Request is always at least 3 bytes
            var authenticationRequest = _stream.ReadChunk(buf => (3 == buf.Count));

            // Only support "no authentication" mode
            if ((0x5 != authenticationRequest[0]) ||
                (0x1 != authenticationRequest[1]) ||
                (0x0 != authenticationRequest[2]))
            {
                throw new NotSupportedException("Malformed or unsupported authentication request from SOCKS 5 client.");
            }

            // Accept "no authentication" mode
            _stream.Write(new byte[] { 0x5, 0x0 });
        }

        /// <summary>
        /// Read the client's connect request.
        /// </summary>
        /// <param name="host">Requested host.</param>
        /// <param name="port">Requested port.</param>
        /// <param name="hostIsDomain">True iff requested host is a domain name (vs. IP).</param>
        public void ReadConnectRequest(out IEnumerable<byte> host, out IEnumerable<byte> port, out bool hostIsDomain)
        {
            // Read the connect request (5 byte header plus content)
            var connectRequest = _stream.ReadChunk(buf =>
                (5 <= buf.Count) &&
                (((0x1 == buf[3]) && (4 + 4 + 2 == buf.Count)) ||
                 ((0x3 == buf[3]) && (4 + 1 + buf[4] + 2 == buf.Count))));

            // Only support TCP/IP stream mode
            if ((0x5 != connectRequest[0]) ||
                (0x1 != connectRequest[1]) ||
                (0x0 != connectRequest[2]))
            {
                throw new NotSupportedException("Malformed or unsupported connect request from SOCKS 5 client.");
            }

            // Extract requested host/port
            hostIsDomain = (0x3 == connectRequest[3]);
            host = hostIsDomain ?
                connectRequest.Skip(5).Take(connectRequest[4]) :
                connectRequest.Skip(4).Take(4);
            port = connectRequest.Skip(connectRequest.Length - 2);
        }

        /// <summary>
        /// Send connect response back to client.
        /// </summary>
        /// <param name="success">True iff successful.</param>
        /// <param name="host">Requested host.</param>
        /// <param name="port">Requested port.</param>
        /// <param name="hostIsDomain">True iff requested host is a domain name (vs. IP).</param>
        public void SendConnectResponse(bool success, IEnumerable<byte> host, IEnumerable<byte> port, bool hostIsDomain)
        {
            // Create response
            var response = new List<byte>();
            response.Add(0x5);
            response.Add(success ? (byte)0x0 : (byte)0x1);
            response.Add(0x0);
            if (hostIsDomain)
            {
#if NOT_SUPPORTED_BY_SOME_APPLICATIONS
                response.Add(0x3);
                response.Add((byte)(host.Count()));
                response.AddRange(host);
            }
            else
#else
                // One application needs an IP address in the response
                // (The SOCKS 4a response doesn't provide the connected-to
                // IP, so return a fake one instead.)
                host = new byte[] { 0x0, 0x0, 0x0, 0x0 };
            }
#endif
            {
                response.Add(0x1);
                response.AddRange(host);
            }
            response.AddRange(port);

            // Send it
            _stream.Write(response);
        }
    }

    /// <summary>
    /// Implements simple SOCKS 4a client behavior for a single connection.
    /// </summary>
    class Socks4Client : SocksBase
    {
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="serverName">Proxy name.</param>
        /// <param name="serverPort">Proxy port.</param>
        public Socks4Client(string proxyName, int proxyPort)
            : base(new TcpClient(proxyName, proxyPort))
        {
        }

        /// <summary>
        /// Issues a connect request and processes the response.
        /// </summary>
        /// <param name="host">Host to connect to.</param>
        /// <param name="port">Port to connect to.</param>
        /// <param name="hostIsDomain">True iff host is a domain name (vs. IP).</param>
        /// <returns></returns>
        public bool Connect(IEnumerable<byte> host, IEnumerable<byte> port, bool hostIsDomain)
        {
            // Create request
            var request = new List<byte>();
            request.Add(0x4);
            request.Add(0x1);
            request.AddRange(port);
            if (hostIsDomain)
            {
                request.AddRange(new byte[] { 0x0, 0x0, 0x0, 0x1 });
                request.Add(0x0);
            }
            request.AddRange(host);
            request.Add(0x0);

            // Send it
            _stream.Write(request);

            // Read the response (8 bytes)
            var _connectResponse = _stream.ReadChunk(buf => (8 == buf.Count));

            // Check response
            if (0x0 != _connectResponse[0])
            {
                throw new NotSupportedException("Malformed or unsupported connect response from SOCKS 4a server.");
            }

            // Return success result
            return (0x5a == _connectResponse[1]);
        }
    }

    /// <summary>
    /// Relays data between two NetworkStream instances.
    /// </summary>
    class NetworkStreamRelay
    {
        private const int BufferSize = 1024;
        private bool _closed;
        private readonly NetworkStream _localStream;
        private readonly NetworkStream _remoteStream;
        private readonly byte[] _localBuffer;
        private readonly byte[] _remoteBuffer;
        private Action _onClose;

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="localStream">Local stream.</param>
        /// <param name="remoteStream">Remote stream.</param>
        /// <param name="onClose">Action to perform when closing the relay.</param>
        public NetworkStreamRelay(NetworkStream localStream, NetworkStream remoteStream, Action onClose)
        {
            _localStream = localStream;
            _remoteStream = remoteStream;
            _localBuffer = new byte[BufferSize];
            _remoteBuffer = new byte[BufferSize];
            _onClose = onClose;

            // Start relaying
            LocalWriteCallback(null);
            RemoteWriteCallback(null);
        }

        /// <summary>
        /// Close the relay.
        /// </summary>
        private void Close()
        {
            _closed = true;
            _localStream.Close();
            _remoteStream.Close();
            _onClose();
        }

        /// <summary>
        /// Handle completion of a read on the local stream.
        /// </summary>
        /// <param name="result">Result.</param>
        private void LocalReadCallback(IAsyncResult result)
        {
            if (!_closed)
            {
                try
                {
                    var bytesRead = _localStream.EndRead(result);
                    if (0 < bytesRead)
                    {
                        _remoteStream.BeginWrite(_localBuffer, 0, bytesRead, LocalWriteCallback, null);
                    }
                    else
                    {
                        Close();
                    }
                }
                catch (IOException)
                {
                    Close();
                }
            }
        }

        /// <summary>
        /// Handle completion of a write on the local stream.
        /// </summary>
        /// <param name="result">Result.</param>
        private void LocalWriteCallback(IAsyncResult result)
        {
            if (!_closed)
            {
                try
                {
                    if (null != result)
                    {
                        _remoteStream.EndWrite(result);
                    }
                    _localStream.BeginRead(_localBuffer, 0, _localBuffer.Length, LocalReadCallback, null);
                }
                catch (IOException)
                {
                    Close();
                }
            }
        }

        /// <summary>
        /// Handle completion of a read on the remote stream.
        /// </summary>
        /// <param name="result">Result.</param>
        private void RemoteReadCallback(IAsyncResult result)
        {
            if (!_closed)
            {
                try
                {
                    var bytesRead = _remoteStream.EndRead(result);
                    if (0 < bytesRead)
                    {
                        _localStream.BeginWrite(_remoteBuffer, 0, bytesRead, RemoteWriteCallback, null);
                    }
                    else
                    {
                        Close();
                    }
                }
                catch (IOException)
                {
                    Close();
                }
            }
        }

        /// <summary>
        /// Handle completion of a write on the remote stream.
        /// </summary>
        /// <param name="result">Result.</param>
        private void RemoteWriteCallback(IAsyncResult result)
        {
            if (!_closed)
            {
                try
                {
                    if (null != result)
                    {
                        _localStream.EndWrite(result);
                    }
                    _remoteStream.BeginRead(_remoteBuffer, 0, _remoteBuffer.Length, RemoteReadCallback, null);
                }
                catch (IOException)
                {
                    Close();
                }
            }
        }
    }

    /// <summary>
    /// Implements helpful extension methods for the NetworkStream class.
    /// </summary>
    static class NetworkStreamExtensions
    {
        /// <summary>
        /// Reads from the stream until a completion condition is met.
        /// </summary>
        /// <param name="self">NetworkStream instance.</param>
        /// <param name="isReadComplete">Function to determine if the read is complete.</param>
        /// <returns>Bytes read.</returns>
        public static byte[] ReadChunk(this NetworkStream self, Func<List<byte>, bool> isReadComplete)
        {
            var buffer = new List<byte>();
            while (!isReadComplete(buffer))
            {
                var readByte = self.ReadByte();
                if (-1 == readByte)
                {
                    throw new IOException("NetworkStream closed before chunk read completed.");
                }
                buffer.Add((byte)readByte);
            }
            return buffer.ToArray();
        }

        /// <summary>
        /// Writes a stream of bytes.
        /// </summary>
        /// <param name="self">NetworkStream instance.</param>
        /// <param name="bytes">Stream of bytes.</param>
        public static void Write(this NetworkStream self, IEnumerable<byte> bytes)
        {
            var buffer = bytes.ToArray();
            self.Write(buffer, 0, buffer.Length);
        }
    }
}
